var jsutils_inspect = require("../jsutils/inspect.js");

var jsutils_keyMap = require("../jsutils/keyMap.js");

var jsutils_printPathArray = require("../jsutils/printPathArray.js");

var type_definition = require("../type/definition.js");

var utilities_coerceInputValue = require("../utilities/coerceInputValue.js");

var utilities_typeFromAST = require("../utilities/typeFromAST.js");

var utilities_valueFromAST = require("../utilities/valueFromAST.js");

var shared = require("../shared.js");

function getArgumentValues(def, node, variableValues) {
  var _node$arguments;
  const coercedValues = {};
  const argumentNodes = null !== (_node$arguments = node.arguments) && void 0 !== _node$arguments ? _node$arguments : [];
  const argNodeMap = jsutils_keyMap.keyMap(argumentNodes, (arg => arg.name.value));
  for (const argDef of def.args) {
    const name = argDef.name;
    const argType = argDef.type;
    const argumentNode = argNodeMap[name];
    if (!argumentNode) {
      if (void 0 !== argDef.defaultValue) {
        coercedValues[name] = argDef.defaultValue;
      } else if (type_definition.isNonNullType(argType)) {
        throw new shared.GraphQLError(`Argument "${name}" of required type "${jsutils_inspect.inspect(argType)}" was not provided.`, {
          nodes: node
        });
      }
      continue;
    }
    const valueNode = argumentNode.value;
    let isNull = valueNode.kind === shared.e.NULL;
    if (valueNode.kind === shared.e.VARIABLE) {
      const variableName = valueNode.name.value;
      if (null == variableValues || !hasOwnProperty(variableValues, variableName)) {
        if (void 0 !== argDef.defaultValue) {
          coercedValues[name] = argDef.defaultValue;
        } else if (type_definition.isNonNullType(argType)) {
          throw new shared.GraphQLError(`Argument "${name}" of required type "${jsutils_inspect.inspect(argType)}" was provided the variable "$${variableName}" which was not provided a runtime value.`, {
            nodes: valueNode
          });
        }
        continue;
      }
      isNull = null == variableValues[variableName];
    }
    if (isNull && type_definition.isNonNullType(argType)) {
      throw new shared.GraphQLError(`Argument "${name}" of non-null type "${jsutils_inspect.inspect(argType)}" must not be null.`, {
        nodes: valueNode
      });
    }
    const coercedValue = utilities_valueFromAST.valueFromAST(valueNode, argType, variableValues);
    if (void 0 === coercedValue) {
      throw new shared.GraphQLError(`Argument "${name}" has invalid value ${shared.print(valueNode)}.`, {
        nodes: valueNode
      });
    }
    coercedValues[name] = coercedValue;
  }
  return coercedValues;
}

function hasOwnProperty(obj, prop) {
  return Object.prototype.hasOwnProperty.call(obj, prop);
}

exports.getArgumentValues = getArgumentValues;

exports.getDirectiveValues = function getDirectiveValues(directiveDef, node, variableValues) {
  var _node$directives;
  const directiveNode = null === (_node$directives = node.directives) || void 0 === _node$directives ? void 0 : _node$directives.find((directive => directive.name.value === directiveDef.name));
  if (directiveNode) {
    return getArgumentValues(directiveDef, directiveNode, variableValues);
  }
};

exports.getVariableValues = function getVariableValues(schema, varDefNodes, inputs, options) {
  const errors = [];
  const maxErrors = null == options ? void 0 : options.maxErrors;
  try {
    const coerced = function coerceVariableValues(schema, varDefNodes, inputs, onError) {
      const coercedValues = {};
      for (const varDefNode of varDefNodes) {
        const varName = varDefNode.variable.name.value;
        const varType = utilities_typeFromAST.typeFromAST(schema, varDefNode.type);
        if (!type_definition.isInputType(varType)) {
          const varTypeStr = shared.print(varDefNode.type);
          onError(new shared.GraphQLError(`Variable "$${varName}" expected value of type "${varTypeStr}" which cannot be used as an input type.`, {
            nodes: varDefNode.type
          }));
          continue;
        }
        if (!hasOwnProperty(inputs, varName)) {
          if (varDefNode.defaultValue) {
            coercedValues[varName] = utilities_valueFromAST.valueFromAST(varDefNode.defaultValue, varType);
          } else if (type_definition.isNonNullType(varType)) {
            const varTypeStr = jsutils_inspect.inspect(varType);
            onError(new shared.GraphQLError(`Variable "$${varName}" of required type "${varTypeStr}" was not provided.`, {
              nodes: varDefNode
            }));
          }
          continue;
        }
        const value = inputs[varName];
        if (null === value && type_definition.isNonNullType(varType)) {
          const varTypeStr = jsutils_inspect.inspect(varType);
          onError(new shared.GraphQLError(`Variable "$${varName}" of non-null type "${varTypeStr}" must not be null.`, {
            nodes: varDefNode
          }));
          continue;
        }
        coercedValues[varName] = utilities_coerceInputValue.coerceInputValue(value, varType, ((path, invalidValue, error) => {
          let prefix = `Variable "$${varName}" got invalid value ` + jsutils_inspect.inspect(invalidValue);
          if (path.length > 0) {
            prefix += ` at "${varName}${jsutils_printPathArray.printPathArray(path)}"`;
          }
          onError(new shared.GraphQLError(prefix + "; " + error.message, {
            nodes: varDefNode,
            originalError: error.originalError
          }));
        }));
      }
      return coercedValues;
    }(schema, varDefNodes, inputs, (error => {
      if (null != maxErrors && errors.length >= maxErrors) {
        throw new shared.GraphQLError("Too many errors processing variables, error limit reached. Execution aborted.");
      }
      errors.push(error);
    }));
    if (0 === errors.length) {
      return {
        coerced
      };
    }
  } catch (error) {
    errors.push(error);
  }
  return {
    errors
  };
};
