var jsutils_inspect = require("../jsutils/inspect.js");

var jsutils_isObjectLike = require("../jsutils/isObjectLike.js");

var type_definition = require("./definition.js");

var shared = require("../shared.js");

const GraphQLInt = new type_definition.GraphQLScalarType({
  name: "Int",
  description: "The `Int` scalar type represents non-fractional signed whole numeric values. Int can represent values between -(2^31) and 2^31 - 1.",
  serialize(outputValue) {
    const coercedValue = serializeObject(outputValue);
    if ("boolean" == typeof coercedValue) {
      return coercedValue ? 1 : 0;
    }
    let num = coercedValue;
    if ("string" == typeof coercedValue && "" !== coercedValue) {
      num = Number(coercedValue);
    }
    if ("number" != typeof num || !Number.isInteger(num)) {
      throw new shared.GraphQLError(`Int cannot represent non-integer value: ${jsutils_inspect.inspect(coercedValue)}`);
    }
    if (num > 2147483647 || num < -2147483648) {
      throw new shared.GraphQLError("Int cannot represent non 32-bit signed integer value: " + jsutils_inspect.inspect(coercedValue));
    }
    return num;
  },
  parseValue(inputValue) {
    if ("number" != typeof inputValue || !Number.isInteger(inputValue)) {
      throw new shared.GraphQLError(`Int cannot represent non-integer value: ${jsutils_inspect.inspect(inputValue)}`);
    }
    if (inputValue > 2147483647 || inputValue < -2147483648) {
      throw new shared.GraphQLError(`Int cannot represent non 32-bit signed integer value: ${inputValue}`);
    }
    return inputValue;
  },
  parseLiteral(valueNode) {
    if (valueNode.kind !== shared.e.INT) {
      throw new shared.GraphQLError(`Int cannot represent non-integer value: ${shared.print(valueNode)}`, {
        nodes: valueNode
      });
    }
    const num = parseInt(valueNode.value, 10);
    if (num > 2147483647 || num < -2147483648) {
      throw new shared.GraphQLError(`Int cannot represent non 32-bit signed integer value: ${valueNode.value}`, {
        nodes: valueNode
      });
    }
    return num;
  }
});

const GraphQLFloat = new type_definition.GraphQLScalarType({
  name: "Float",
  description: "The `Float` scalar type represents signed double-precision fractional values as specified by [IEEE 754](https://en.wikipedia.org/wiki/IEEE_floating_point).",
  serialize(outputValue) {
    const coercedValue = serializeObject(outputValue);
    if ("boolean" == typeof coercedValue) {
      return coercedValue ? 1 : 0;
    }
    let num = coercedValue;
    if ("string" == typeof coercedValue && "" !== coercedValue) {
      num = Number(coercedValue);
    }
    if ("number" != typeof num || !Number.isFinite(num)) {
      throw new shared.GraphQLError(`Float cannot represent non numeric value: ${jsutils_inspect.inspect(coercedValue)}`);
    }
    return num;
  },
  parseValue(inputValue) {
    if ("number" != typeof inputValue || !Number.isFinite(inputValue)) {
      throw new shared.GraphQLError(`Float cannot represent non numeric value: ${jsutils_inspect.inspect(inputValue)}`);
    }
    return inputValue;
  },
  parseLiteral(valueNode) {
    if (valueNode.kind !== shared.e.FLOAT && valueNode.kind !== shared.e.INT) {
      throw new shared.GraphQLError(`Float cannot represent non numeric value: ${shared.print(valueNode)}`, valueNode);
    }
    return parseFloat(valueNode.value);
  }
});

const GraphQLString = new type_definition.GraphQLScalarType({
  name: "String",
  description: "The `String` scalar type represents textual data, represented as UTF-8 character sequences. The String type is most often used by GraphQL to represent free-form human-readable text.",
  serialize(outputValue) {
    const coercedValue = serializeObject(outputValue);
    if ("string" == typeof coercedValue) {
      return coercedValue;
    }
    if ("boolean" == typeof coercedValue) {
      return coercedValue ? "true" : "false";
    }
    if ("number" == typeof coercedValue && Number.isFinite(coercedValue)) {
      return coercedValue.toString();
    }
    throw new shared.GraphQLError(`String cannot represent value: ${jsutils_inspect.inspect(outputValue)}`);
  },
  parseValue(inputValue) {
    if ("string" != typeof inputValue) {
      throw new shared.GraphQLError(`String cannot represent a non string value: ${jsutils_inspect.inspect(inputValue)}`);
    }
    return inputValue;
  },
  parseLiteral(valueNode) {
    if (valueNode.kind !== shared.e.STRING) {
      throw new shared.GraphQLError(`String cannot represent a non string value: ${shared.print(valueNode)}`, {
        nodes: valueNode
      });
    }
    return valueNode.value;
  }
});

const GraphQLBoolean = new type_definition.GraphQLScalarType({
  name: "Boolean",
  description: "The `Boolean` scalar type represents `true` or `false`.",
  serialize(outputValue) {
    const coercedValue = serializeObject(outputValue);
    if ("boolean" == typeof coercedValue) {
      return coercedValue;
    }
    if (Number.isFinite(coercedValue)) {
      return 0 !== coercedValue;
    }
    throw new shared.GraphQLError(`Boolean cannot represent a non boolean value: ${jsutils_inspect.inspect(coercedValue)}`);
  },
  parseValue(inputValue) {
    if ("boolean" != typeof inputValue) {
      throw new shared.GraphQLError(`Boolean cannot represent a non boolean value: ${jsutils_inspect.inspect(inputValue)}`);
    }
    return inputValue;
  },
  parseLiteral(valueNode) {
    if (valueNode.kind !== shared.e.BOOLEAN) {
      throw new shared.GraphQLError(`Boolean cannot represent a non boolean value: ${shared.print(valueNode)}`, {
        nodes: valueNode
      });
    }
    return valueNode.value;
  }
});

const GraphQLID = new type_definition.GraphQLScalarType({
  name: "ID",
  description: 'The `ID` scalar type represents a unique identifier, often used to refetch an object or as key for a cache. The ID type appears in a JSON response as a String; however, it is not intended to be human-readable. When expected as an input type, any string (such as `"4"`) or integer (such as `4`) input value will be accepted as an ID.',
  serialize(outputValue) {
    const coercedValue = serializeObject(outputValue);
    if ("string" == typeof coercedValue) {
      return coercedValue;
    }
    if (Number.isInteger(coercedValue)) {
      return String(coercedValue);
    }
    throw new shared.GraphQLError(`ID cannot represent value: ${jsutils_inspect.inspect(outputValue)}`);
  },
  parseValue(inputValue) {
    if ("string" == typeof inputValue) {
      return inputValue;
    }
    if ("number" == typeof inputValue && Number.isInteger(inputValue)) {
      return inputValue.toString();
    }
    throw new shared.GraphQLError(`ID cannot represent value: ${jsutils_inspect.inspect(inputValue)}`);
  },
  parseLiteral(valueNode) {
    if (valueNode.kind !== shared.e.STRING && valueNode.kind !== shared.e.INT) {
      throw new shared.GraphQLError("ID cannot represent a non-string and non-integer value: " + shared.print(valueNode), {
        nodes: valueNode
      });
    }
    return valueNode.value;
  }
});

const specifiedScalarTypes = [ GraphQLString, GraphQLInt, GraphQLFloat, GraphQLBoolean, GraphQLID ];

function serializeObject(outputValue) {
  if (jsutils_isObjectLike.isObjectLike(outputValue)) {
    if ("function" == typeof outputValue.valueOf) {
      const valueOfResult = outputValue.valueOf();
      if (!jsutils_isObjectLike.isObjectLike(valueOfResult)) {
        return valueOfResult;
      }
    }
    if ("function" == typeof outputValue.toJSON) {
      return outputValue.toJSON();
    }
  }
  return outputValue;
}

exports.GraphQLBoolean = GraphQLBoolean;

exports.GraphQLFloat = GraphQLFloat;

exports.GraphQLID = GraphQLID;

exports.GraphQLInt = GraphQLInt;

exports.GraphQLString = GraphQLString;

exports.isSpecifiedScalarType = function isSpecifiedScalarType(type) {
  return specifiedScalarTypes.some((({name}) => type.name === name));
};

exports.specifiedScalarTypes = specifiedScalarTypes;
