import { inspect } from "../jsutils/inspect.mjs";

import { OperationTypeNode } from "../language/ast.mjs";

import { isTypeSubTypeOf, isEqualType } from "../utilities/typeComparators.mjs";

import { isObjectType, isInputType, isRequiredArgument, isNamedType, isInterfaceType, isUnionType, isEnumType, isInputObjectType, isOutputType, isRequiredInputField, isNonNullType } from "./definition.mjs";

import { isDirective, GraphQLDeprecatedDirective } from "./directives.mjs";

import { isIntrospectionType } from "./introspection.mjs";

import { assertSchema } from "./schema.mjs";

import { GraphQLError } from "../shared.mjs";

function validateSchema(schema) {
  assertSchema(schema);
  if (schema.__validationErrors) {
    return schema.__validationErrors;
  }
  const context = new SchemaValidationContext(schema);
  !function validateRootTypes(context) {
    const schema = context.schema;
    const queryType = schema.getQueryType();
    if (!queryType) {
      context.reportError("Query root type must be provided.", schema.astNode);
    } else if (!isObjectType(queryType)) {
      var _getOperationTypeNode;
      context.reportError(`Query root type must be Object type, it cannot be ${inspect(queryType)}.`, null !== (_getOperationTypeNode = getOperationTypeNode(schema, OperationTypeNode.QUERY)) && void 0 !== _getOperationTypeNode ? _getOperationTypeNode : queryType.astNode);
    }
    const mutationType = schema.getMutationType();
    if (mutationType && !isObjectType(mutationType)) {
      var _getOperationTypeNode2;
      context.reportError(`Mutation root type must be Object type if provided, it cannot be ${inspect(mutationType)}.`, null !== (_getOperationTypeNode2 = getOperationTypeNode(schema, OperationTypeNode.MUTATION)) && void 0 !== _getOperationTypeNode2 ? _getOperationTypeNode2 : mutationType.astNode);
    }
    const subscriptionType = schema.getSubscriptionType();
    if (subscriptionType && !isObjectType(subscriptionType)) {
      var _getOperationTypeNode3;
      context.reportError(`Subscription root type must be Object type if provided, it cannot be ${inspect(subscriptionType)}.`, null !== (_getOperationTypeNode3 = getOperationTypeNode(schema, OperationTypeNode.SUBSCRIPTION)) && void 0 !== _getOperationTypeNode3 ? _getOperationTypeNode3 : subscriptionType.astNode);
    }
  }(context);
  !function validateDirectives(context) {
    for (const directive of context.schema.getDirectives()) {
      if (!isDirective(directive)) {
        context.reportError(`Expected directive but got: ${inspect(directive)}.`, null == directive ? void 0 : directive.astNode);
        continue;
      }
      validateName(context, directive);
      for (const arg of directive.args) {
        validateName(context, arg);
        if (!isInputType(arg.type)) {
          context.reportError(`The type of @${directive.name}(${arg.name}:) must be Input Type but got: ${inspect(arg.type)}.`, arg.astNode);
        }
        if (isRequiredArgument(arg) && null != arg.deprecationReason) {
          var _arg$astNode;
          context.reportError(`Required argument @${directive.name}(${arg.name}:) cannot be deprecated.`, [ getDeprecatedDirectiveNode(arg.astNode), null === (_arg$astNode = arg.astNode) || void 0 === _arg$astNode ? void 0 : _arg$astNode.type ]);
        }
      }
    }
  }(context);
  !function validateTypes(context) {
    const validateInputObjectCircularRefs = function createInputObjectCircularRefsValidator(context) {
      const visitedTypes = Object.create(null);
      const fieldPath = [];
      const fieldPathIndexByTypeName = Object.create(null);
      return detectCycleRecursive;
      function detectCycleRecursive(inputObj) {
        if (visitedTypes[inputObj.name]) {
          return;
        }
        visitedTypes[inputObj.name] = !0;
        fieldPathIndexByTypeName[inputObj.name] = fieldPath.length;
        const fields = Object.values(inputObj.getFields());
        for (const field of fields) {
          if (isNonNullType(field.type) && isInputObjectType(field.type.ofType)) {
            const fieldType = field.type.ofType;
            const cycleIndex = fieldPathIndexByTypeName[fieldType.name];
            fieldPath.push(field);
            if (void 0 === cycleIndex) {
              detectCycleRecursive(fieldType);
            } else {
              const cyclePath = fieldPath.slice(cycleIndex);
              const pathStr = cyclePath.map((fieldObj => fieldObj.name)).join(".");
              context.reportError(`Cannot reference Input Object "${fieldType.name}" within itself through a series of non-null fields: "${pathStr}".`, cyclePath.map((fieldObj => fieldObj.astNode)));
            }
            fieldPath.pop();
          }
        }
        fieldPathIndexByTypeName[inputObj.name] = void 0;
      }
    }(context);
    const typeMap = context.schema.getTypeMap();
    for (const type of Object.values(typeMap)) {
      if (!isNamedType(type)) {
        context.reportError(`Expected GraphQL named type but got: ${inspect(type)}.`, type.astNode);
        continue;
      }
      if (!isIntrospectionType(type)) {
        validateName(context, type);
      }
      if (isObjectType(type)) {
        validateFields(context, type);
        validateInterfaces(context, type);
      } else if (isInterfaceType(type)) {
        validateFields(context, type);
        validateInterfaces(context, type);
      } else if (isUnionType(type)) {
        validateUnionMembers(context, type);
      } else if (isEnumType(type)) {
        validateEnumValues(context, type);
      } else if (isInputObjectType(type)) {
        validateInputFields(context, type);
        validateInputObjectCircularRefs(type);
      }
    }
  }(context);
  const errors = context.getErrors();
  schema.__validationErrors = errors;
  return errors;
}

function assertValidSchema(schema) {
  const errors = validateSchema(schema);
  if (0 !== errors.length) {
    throw new Error(errors.map((error => error.message)).join("\n\n"));
  }
}

class SchemaValidationContext {
  constructor(schema) {
    this._errors = [];
    this.schema = schema;
  }
  reportError(message, nodes) {
    const _nodes = Array.isArray(nodes) ? nodes.filter(Boolean) : nodes;
    this._errors.push(new GraphQLError(message, {
      nodes: _nodes
    }));
  }
  getErrors() {
    return this._errors;
  }
}

function getOperationTypeNode(schema, operation) {
  var _flatMap$find;
  return null === (_flatMap$find = [ schema.astNode, ...schema.extensionASTNodes ].flatMap((schemaNode => {
    var _schemaNode$operation;
    return null !== (_schemaNode$operation = null == schemaNode ? void 0 : schemaNode.operationTypes) && void 0 !== _schemaNode$operation ? _schemaNode$operation : [];
  })).find((operationNode => operationNode.operation === operation))) || void 0 === _flatMap$find ? void 0 : _flatMap$find.type;
}

function validateName(context, node) {
  if (node.name.startsWith("__")) {
    context.reportError(`Name "${node.name}" must not begin with "__", which is reserved by GraphQL introspection.`, node.astNode);
  }
}

function validateFields(context, type) {
  const fields = Object.values(type.getFields());
  if (0 === fields.length) {
    context.reportError(`Type ${type.name} must define one or more fields.`, [ type.astNode, ...type.extensionASTNodes ]);
  }
  for (const field of fields) {
    validateName(context, field);
    if (!isOutputType(field.type)) {
      var _field$astNode;
      context.reportError(`The type of ${type.name}.${field.name} must be Output Type but got: ${inspect(field.type)}.`, null === (_field$astNode = field.astNode) || void 0 === _field$astNode ? void 0 : _field$astNode.type);
    }
    for (const arg of field.args) {
      const argName = arg.name;
      validateName(context, arg);
      if (!isInputType(arg.type)) {
        var _arg$astNode2;
        context.reportError(`The type of ${type.name}.${field.name}(${argName}:) must be Input Type but got: ${inspect(arg.type)}.`, null === (_arg$astNode2 = arg.astNode) || void 0 === _arg$astNode2 ? void 0 : _arg$astNode2.type);
      }
      if (isRequiredArgument(arg) && null != arg.deprecationReason) {
        var _arg$astNode3;
        context.reportError(`Required argument ${type.name}.${field.name}(${argName}:) cannot be deprecated.`, [ getDeprecatedDirectiveNode(arg.astNode), null === (_arg$astNode3 = arg.astNode) || void 0 === _arg$astNode3 ? void 0 : _arg$astNode3.type ]);
      }
    }
  }
}

function validateInterfaces(context, type) {
  const ifaceTypeNames = Object.create(null);
  for (const iface of type.getInterfaces()) {
    if (!isInterfaceType(iface)) {
      context.reportError(`Type ${inspect(type)} must only implement Interface types, it cannot implement ${inspect(iface)}.`, getAllImplementsInterfaceNodes(type, iface));
      continue;
    }
    if (type === iface) {
      context.reportError(`Type ${type.name} cannot implement itself because it would create a circular reference.`, getAllImplementsInterfaceNodes(type, iface));
      continue;
    }
    if (ifaceTypeNames[iface.name]) {
      context.reportError(`Type ${type.name} can only implement ${iface.name} once.`, getAllImplementsInterfaceNodes(type, iface));
      continue;
    }
    ifaceTypeNames[iface.name] = !0;
    validateTypeImplementsAncestors(context, type, iface);
    validateTypeImplementsInterface(context, type, iface);
  }
}

function validateTypeImplementsInterface(context, type, iface) {
  const typeFieldMap = type.getFields();
  for (const ifaceField of Object.values(iface.getFields())) {
    const fieldName = ifaceField.name;
    const typeField = typeFieldMap[fieldName];
    if (!typeField) {
      context.reportError(`Interface field ${iface.name}.${fieldName} expected but ${type.name} does not provide it.`, [ ifaceField.astNode, type.astNode, ...type.extensionASTNodes ]);
      continue;
    }
    if (!isTypeSubTypeOf(context.schema, typeField.type, ifaceField.type)) {
      var _ifaceField$astNode, _typeField$astNode;
      context.reportError(`Interface field ${iface.name}.${fieldName} expects type ${inspect(ifaceField.type)} but ${type.name}.${fieldName} is type ${inspect(typeField.type)}.`, [ null === (_ifaceField$astNode = ifaceField.astNode) || void 0 === _ifaceField$astNode ? void 0 : _ifaceField$astNode.type, null === (_typeField$astNode = typeField.astNode) || void 0 === _typeField$astNode ? void 0 : _typeField$astNode.type ]);
    }
    for (const ifaceArg of ifaceField.args) {
      const argName = ifaceArg.name;
      const typeArg = typeField.args.find((arg => arg.name === argName));
      if (!typeArg) {
        context.reportError(`Interface field argument ${iface.name}.${fieldName}(${argName}:) expected but ${type.name}.${fieldName} does not provide it.`, [ ifaceArg.astNode, typeField.astNode ]);
        continue;
      }
      if (!isEqualType(ifaceArg.type, typeArg.type)) {
        var _ifaceArg$astNode, _typeArg$astNode;
        context.reportError(`Interface field argument ${iface.name}.${fieldName}(${argName}:) expects type ${inspect(ifaceArg.type)} but ${type.name}.${fieldName}(${argName}:) is type ${inspect(typeArg.type)}.`, [ null === (_ifaceArg$astNode = ifaceArg.astNode) || void 0 === _ifaceArg$astNode ? void 0 : _ifaceArg$astNode.type, null === (_typeArg$astNode = typeArg.astNode) || void 0 === _typeArg$astNode ? void 0 : _typeArg$astNode.type ]);
      }
    }
    for (const typeArg of typeField.args) {
      const argName = typeArg.name;
      if (!ifaceField.args.find((arg => arg.name === argName)) && isRequiredArgument(typeArg)) {
        context.reportError(`Object field ${type.name}.${fieldName} includes required argument ${argName} that is missing from the Interface field ${iface.name}.${fieldName}.`, [ typeArg.astNode, ifaceField.astNode ]);
      }
    }
  }
}

function validateTypeImplementsAncestors(context, type, iface) {
  const ifaceInterfaces = type.getInterfaces();
  for (const transitive of iface.getInterfaces()) {
    if (!ifaceInterfaces.includes(transitive)) {
      context.reportError(transitive === type ? `Type ${type.name} cannot implement ${iface.name} because it would create a circular reference.` : `Type ${type.name} must implement ${transitive.name} because it is implemented by ${iface.name}.`, [ ...getAllImplementsInterfaceNodes(iface, transitive), ...getAllImplementsInterfaceNodes(type, iface) ]);
    }
  }
}

function validateUnionMembers(context, union) {
  const memberTypes = union.getTypes();
  if (0 === memberTypes.length) {
    context.reportError(`Union type ${union.name} must define one or more member types.`, [ union.astNode, ...union.extensionASTNodes ]);
  }
  const includedTypeNames = Object.create(null);
  for (const memberType of memberTypes) {
    if (includedTypeNames[memberType.name]) {
      context.reportError(`Union type ${union.name} can only include type ${memberType.name} once.`, getUnionMemberTypeNodes(union, memberType.name));
      continue;
    }
    includedTypeNames[memberType.name] = !0;
    if (!isObjectType(memberType)) {
      context.reportError(`Union type ${union.name} can only include Object types, it cannot include ${inspect(memberType)}.`, getUnionMemberTypeNodes(union, String(memberType)));
    }
  }
}

function validateEnumValues(context, enumType) {
  const enumValues = enumType.getValues();
  if (0 === enumValues.length) {
    context.reportError(`Enum type ${enumType.name} must define one or more values.`, [ enumType.astNode, ...enumType.extensionASTNodes ]);
  }
  for (const enumValue of enumValues) {
    validateName(context, enumValue);
  }
}

function validateInputFields(context, inputObj) {
  const fields = Object.values(inputObj.getFields());
  if (0 === fields.length) {
    context.reportError(`Input Object type ${inputObj.name} must define one or more fields.`, [ inputObj.astNode, ...inputObj.extensionASTNodes ]);
  }
  for (const field of fields) {
    validateName(context, field);
    if (!isInputType(field.type)) {
      var _field$astNode2;
      context.reportError(`The type of ${inputObj.name}.${field.name} must be Input Type but got: ${inspect(field.type)}.`, null === (_field$astNode2 = field.astNode) || void 0 === _field$astNode2 ? void 0 : _field$astNode2.type);
    }
    if (isRequiredInputField(field) && null != field.deprecationReason) {
      var _field$astNode3;
      context.reportError(`Required input field ${inputObj.name}.${field.name} cannot be deprecated.`, [ getDeprecatedDirectiveNode(field.astNode), null === (_field$astNode3 = field.astNode) || void 0 === _field$astNode3 ? void 0 : _field$astNode3.type ]);
    }
  }
}

function getAllImplementsInterfaceNodes(type, iface) {
  const {astNode, extensionASTNodes} = type;
  return (null != astNode ? [ astNode, ...extensionASTNodes ] : extensionASTNodes).flatMap((typeNode => {
    var _typeNode$interfaces;
    return null !== (_typeNode$interfaces = typeNode.interfaces) && void 0 !== _typeNode$interfaces ? _typeNode$interfaces : [];
  })).filter((ifaceNode => ifaceNode.name.value === iface.name));
}

function getUnionMemberTypeNodes(union, typeName) {
  const {astNode, extensionASTNodes} = union;
  return (null != astNode ? [ astNode, ...extensionASTNodes ] : extensionASTNodes).flatMap((unionNode => {
    var _unionNode$types;
    return null !== (_unionNode$types = unionNode.types) && void 0 !== _unionNode$types ? _unionNode$types : [];
  })).filter((typeNode => typeNode.name.value === typeName));
}

function getDeprecatedDirectiveNode(definitionNode) {
  var _definitionNode$direc;
  return null == definitionNode ? void 0 : null === (_definitionNode$direc = definitionNode.directives) || void 0 === _definitionNode$direc ? void 0 : _definitionNode$direc.find((node => node.name.value === GraphQLDeprecatedDirective.name));
}

export { assertValidSchema, validateSchema };
