var jsutils_invariant = require("../jsutils/invariant.js");

var jsutils_keyValMap = require("../jsutils/keyValMap.js");

var jsutils_naturalCompare = require("../jsutils/naturalCompare.js");

var type_definition = require("../type/definition.js");

var type_directives = require("../type/directives.js");

var type_introspection = require("../type/introspection.js");

var type_schema = require("../type/schema.js");

function sortObjMap(map, sortValueFn) {
  const sortedMap = Object.create(null);
  for (const key of Object.keys(map).sort(jsutils_naturalCompare.naturalCompare)) {
    sortedMap[key] = sortValueFn(map[key]);
  }
  return sortedMap;
}

function sortByName(array) {
  return sortBy(array, (obj => obj.name));
}

function sortBy(array, mapToKey) {
  return array.slice().sort(((obj1, obj2) => {
    const key1 = mapToKey(obj1);
    const key2 = mapToKey(obj2);
    return jsutils_naturalCompare.naturalCompare(key1, key2);
  }));
}

exports.lexicographicSortSchema = function lexicographicSortSchema(schema) {
  const schemaConfig = schema.toConfig();
  const typeMap = jsutils_keyValMap.keyValMap(sortByName(schemaConfig.types), (type => type.name), (function sortNamedType(type) {
    if (type_definition.isScalarType(type) || type_introspection.isIntrospectionType(type)) {
      return type;
    }
    if (type_definition.isObjectType(type)) {
      const config = type.toConfig();
      return new type_definition.GraphQLObjectType({
        ...config,
        interfaces: () => sortTypes(config.interfaces),
        fields: () => sortFields(config.fields)
      });
    }
    if (type_definition.isInterfaceType(type)) {
      const config = type.toConfig();
      return new type_definition.GraphQLInterfaceType({
        ...config,
        interfaces: () => sortTypes(config.interfaces),
        fields: () => sortFields(config.fields)
      });
    }
    if (type_definition.isUnionType(type)) {
      const config = type.toConfig();
      return new type_definition.GraphQLUnionType({
        ...config,
        types: () => sortTypes(config.types)
      });
    }
    if (type_definition.isEnumType(type)) {
      const config = type.toConfig();
      return new type_definition.GraphQLEnumType({
        ...config,
        values: sortObjMap(config.values, (value => value))
      });
    }
    if (type_definition.isInputObjectType(type)) {
      const config = type.toConfig();
      return new type_definition.GraphQLInputObjectType({
        ...config,
        fields: () => function sortInputFields(fieldsMap) {
          return sortObjMap(fieldsMap, (field => ({
            ...field,
            type: replaceType(field.type)
          })));
        }(config.fields)
      });
    }
    jsutils_invariant.invariant(!1);
  }));
  return new type_schema.GraphQLSchema({
    ...schemaConfig,
    types: Object.values(typeMap),
    directives: sortByName(schemaConfig.directives).map((function sortDirective(directive) {
      const config = directive.toConfig();
      return new type_directives.GraphQLDirective({
        ...config,
        locations: sortBy(config.locations, (x => x)),
        args: sortArgs(config.args)
      });
    })),
    query: replaceMaybeType(schemaConfig.query),
    mutation: replaceMaybeType(schemaConfig.mutation),
    subscription: replaceMaybeType(schemaConfig.subscription)
  });
  function replaceType(type) {
    if (type_definition.isListType(type)) {
      return new type_definition.GraphQLList(replaceType(type.ofType));
    } else if (type_definition.isNonNullType(type)) {
      return new type_definition.GraphQLNonNull(replaceType(type.ofType));
    }
    return replaceNamedType(type);
  }
  function replaceNamedType(type) {
    return typeMap[type.name];
  }
  function replaceMaybeType(maybeType) {
    return maybeType && replaceNamedType(maybeType);
  }
  function sortArgs(args) {
    return sortObjMap(args, (arg => ({
      ...arg,
      type: replaceType(arg.type)
    })));
  }
  function sortFields(fieldsMap) {
    return sortObjMap(fieldsMap, (field => ({
      ...field,
      type: replaceType(field.type),
      args: field.args && sortArgs(field.args)
    })));
  }
  function sortTypes(array) {
    return sortByName(array).map(replaceNamedType);
  }
};
